//! Server-only features

use persistence::Persistent;
use state::PersistentState;
use tracing::level_filters::LevelFilter;
use type_toppings::ResultExt;

pub mod cli {
    //! Server CLI interface

    #[derive(clap::Parser)]
    #[command(version, about, long_about = None)]
    pub struct Cli {
        #[command(subcommand)]
        pub command: Command,
    }

    #[derive(clap::Subcommand, Default)]
    pub enum Command {
        #[default]
        Serve,

        /// Resets state, replacing it with defaults
        ResetState,
    }
}

pub mod state {
    //! Server state

    use super::persistence::Persistent;
    use serde::Deserialize;
    use serde::Serialize;
    use std::collections::BTreeMap;
    use std::collections::BTreeSet;

    #[derive(Clone, Debug)]
    pub struct State {
        pub persistent: Persistent<PersistentState>,
    }

    #[derive(Serialize, Deserialize, Clone, Debug, Default)]
    pub struct PersistentState {
        pub wall: Wall,
    }

    #[derive(Serialize, Deserialize, Clone, Debug, Default)]
    pub struct Wall {
        pub rows: u32,
        pub cols: u32,
        pub holds: BTreeMap<HoldPosition, Hold>,
        pub routes: BTreeSet<route::Route>,
    }

    #[derive(Serialize, Deserialize, Clone, Debug, Default, PartialEq, Eq, Hash, PartialOrd, Ord)]
    pub struct HoldPosition {
        pub row: u32,
        pub col: u32,
    }

    #[derive(Serialize, Deserialize, Clone, Debug, Default)]
    pub struct Hold {
        pub position: HoldPosition,
        pub image: Option<Image>,
    }

    #[derive(Serialize, Deserialize, Clone, Debug, Default)]
    pub struct Image {}

    mod route {
        use super::HoldPosition;
        use serde::Deserialize;
        use serde::Serialize;
        use std::collections::BTreeMap;

        #[derive(Serialize, Deserialize, Clone, Debug, Default, PartialEq, Eq, Hash, PartialOrd, Ord)]
        pub struct Route {
            pub holds: BTreeMap<HoldPosition, HoldRole>,
        }

        /// The role of a hold on a route
        #[derive(Serialize, Deserialize, Clone, Debug, Copy, PartialEq, Eq, Hash, PartialOrd, Ord)]
        pub enum HoldRole {
            /// Start hold
            Start,

            /// Any hold on the route without a specific role
            Normal,

            /// Zone hold
            Zone,

            /// End hold
            End,
        }
    }
}

pub mod persistence;

pub const STATE_FILE: &str = "state.json";

#[tracing::instrument]
pub async fn main() {
    use crate::server::cli::Cli;
    use crate::server::cli::Command;
    use clap::Parser as _;
    use tracing_subscriber::EnvFilter;

    tracing_subscriber::fmt()
        .without_time()
        .with_file(true)
        .with_line_number(true)
        .with_target(false)
        .with_ansi(true)
        .with_env_filter(EnvFilter::builder().with_default_directive(LevelFilter::DEBUG.into()).from_env_lossy())
        .pretty()
        .init();

    let cli = Cli::parse();
    match cli.command {
        Command::Serve => serve().await.unwrap_or_report(),
        Command::ResetState => {
            let s = PersistentState::default();
            let p = camino::Utf8Path::new(STATE_FILE);
            tracing::info!("Resetting state to default: {p}");
            Persistent::persist(p, &s).await.unwrap_or_report();
        }
    }
}

#[tracing::instrument(err)]
async fn serve() -> Result<(), Error> {
    use crate::app::App;
    use crate::app::shell;
    use crate::server::state::State;
    use axum::Router;
    use leptos::prelude::*;
    use leptos_axum::LeptosRoutes;
    use leptos_axum::generate_route_list;

    // Setting get_configuration(None) means we'll be using cargo-leptos's env values
    // For deployment these variables are:
    // <https://github.com/leptos-rs/start-axum#executing-a-server-on-a-remote-machine-without-the-toolchain>
    // Alternately a file can be specified such as Some("Cargo.toml")
    // The file would need to be included with the executable when moved to deployment
    let conf = get_configuration(None).unwrap_or_report();
    let leptos_options = conf.leptos_options;
    let addr = leptos_options.site_addr;
    let routes = generate_route_list(App);

    tracing::info!("Loading state");
    let server_state = State {
        persistent: Persistent::<PersistentState>::load(STATE_FILE.into()).await?,
    };

    // build our application with a route
    let app = Router::new()
        .leptos_routes_with_context(&leptos_options, routes, move || provide_context(server_state.clone()), {
            let leptos_options = leptos_options.clone();
            move || shell(leptos_options.clone())
        })
        .fallback(leptos_axum::file_and_error_handler(shell))
        .with_state(leptos_options);

    let listener = tokio::net::TcpListener::bind(&addr).await?;
    tracing::info!("Listening on http://{addr}");
    axum::serve(listener, app.into_make_service()).await?;

    Ok(())
}

#[derive(Debug, derive_more::Error, derive_more::Display, derive_more::From)]
#[display("Server crash")]
pub enum Error {
    Io(std::io::Error),
    Persistence(persistence::Error),
}
